--[[
	GuildBankSpoofer attempts to fool addons into thinking your guild bank is your personal bank.
	I use Reagent Restocker's feature to overstock items to bank. I'd like it to move overstocked items to guild bank and withdraw them when they're needed in my bag.
]]

local folder, core = ...
_GBS = core --needs to be made global

--[===[@debug@
--@end-debug@]===]
local LibStub = LibStub
local _G = _G
local chatFrame
local pairs = pairs
local tostring = tostring
local table_getn = table.getn
local InterfaceOptionsFrame_OpenToCategory = InterfaceOptionsFrame_OpenToCategory
local _
local GetItemInfo = GetItemInfo
local GetNumGuildBankTabs = GetNumGuildBankTabs
local ipairs = ipairs
local GetFramesRegisteredForEvent = GetFramesRegisteredForEvent
local table_concat = table.concat
local GetCursorInfo = GetCursorInfo
local CreateFrame = CreateFrame
local strsplit = strsplit
local tonumber = tonumber
local select = select
local BANK_CONTAINER = BANK_CONTAINER
local UnitIsUnit = UnitIsUnit


core.title		= "Guild Bank Spoofer"
core.version	= GetAddOnMetadata(folder, "X-Curse-Packaged-Version") or ""
core.titleFull	= core.title.." "..core.version
core.addonDir = "Interface\\AddOns\\"..folder.."\\"

LibStub("AceAddon-3.0"):NewAddon(core, folder, "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0", "AceTimer-3.0") 

core.L = LibStub("AceLocale-3.0"):GetLocale(folder, true)
local L = core.L

local P --db.profile
core.defaultSettings = {
	profile = {
	},
}

local spoofing = false --User has clicked spoofing button. Our events won't fake bank events unless this is enabled.

local slotsPerBankTab = 98 --slots per tab.
local fakeBagID = 51809 --Portable Hole

local unmovedLockDelay = 3 --If a slot changes from locked to unlocked, wait 3 seconds before firing locked event. If GUILDBANK_ITEM_LOCK_CHANGED fires before that time we cancel the timer.
--[[ unmovedLockDelay
	When we pickup a item in the guild bank, the slot becomes locked. When we move the item to a new slot the old slot becomes unlocked. Then 1-2 seconds later the server tells us the item has been moved.
	If the user drops a item, we wait 3 seconds for GUILDBANK_ITEM_LOCK_CHANGED to fire with the new slot location which we'll fire unlock event for. If 3 second passes, then we'll fire the unlock event with the previous slot.
]]

local regEvents = {
--~ 	"ACTIONBAR_SLOT_CHANGED",
	"GUILDBANKFRAME_OPENED",
	"GUILDBANKFRAME_CLOSED",
	"BAG_UPDATE",
--~ 	"PLAYERBANKBAGSLOTS_CHANGED",--the bag slots move, not the items in the bag.
	"PLAYERBANKSLOTS_CHANGED",
	"GUILDBANKBAGSLOTS_CHANGED",
	"ITEM_LOCK_CHANGED",
	"GUILDBANK_ITEM_LOCK_CHANGED",
}
local coreOpts

function core:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("GBS_DB", core.defaultSettings)
--~ 	self.CoreOptionsTable.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileDeleted", "OnProfileChanged")
	self:RegisterChatCommand("gbs", "MySlashProcessorFunc")
	
	chatFrame = _G["ChatFrame1"]
	
	--[[
	local config = LibStub("AceConfig-3.0")
	local dialog = LibStub("AceConfigDialog-3.0")
	config:RegisterOptionsTable(folder, self.CoreOptionsTable)
	coreOpts = dialog:AddToBlizOptions(folder, self.titleFull)
	]]
	
	self:RawHook("GetContainerNumSlots", true)
	self:RawHook("GetContainerItemLink", true)
	self:RawHook("GetContainerItemInfo", true)
	self:RawHook("SplitContainerItem", true)
	self:RawHook("PickupContainerItem", true)
	self:RawHook("PutItemInBag", true)
	self:RawHook("GetItemCount", true)
	self:RawHook("GetNumBankSlots", true)
	
	self:RawHook("GetInventoryItemLink", true)
	self:RawHook("GetInventoryItemTexture", true)
	
	
--~ 	self:RawHook("UseContainerItem", true)

end



----------------------------------------------------------------------
function core:OnProfileChanged(...)									--
-- User has reset proflie, so we reset our spell exists options.	--
----------------------------------------------------------------------
	-- Shut down anything left from previous settings
	self:Disable()
	-- Enable again with the new settings
	self:Enable()
end

function core:OnEnable()
	P = self.db.profile

	for i, event in pairs(regEvents) do 
		self:RegisterEvent(event)
	end
end



function core:OnDisable()
--~ 	timerFrame:SetScript("OnUpdate", nil)
end


local DEBUG = false
--[===[@debug@
DEBUG = true
--@end-debug@]===]

local strWhiteBar		= "|cffffff00 || |r" -- a white bar to seperate the debug info.
local colouredName		= "|cff7f7f7f{|r|cffff0000GBS|r|cff7f7f7f}|r "
local function echo(...)
	local tbl  = {...}
	local msg = tostring(tbl[1])
	for i=2,table_getn(tbl) do 
		msg = msg..strWhiteBar..tostring(tbl[i])
	end
	

	local cf = chatFrame
	if cf then
		cf:AddMessage(colouredName..msg,.7,.7,.7)
	end
end
core.echo = echo

--~ local whiteText			= "|cffffffff%s|r"
local strDebugFrom		= "|cffffff00[%s]|r" --Yellow function name. help pinpoint where the debug msg is from.
-----------------------------
local function Debug(from, ...)	--
-- simple print function.	--
------------------------------
	if DEBUG == false then
		return 
	end
	local tbl  = {...}
	local msg = tostring(tbl[1])
	for i=2,table_getn(tbl) do 
		msg = msg..strWhiteBar..tostring(tbl[i])
	end

	
	echo(strDebugFrom:format(from).." "..msg)
end
core.Debug = Debug

----------------------------------------------
function core:MySlashProcessorFunc(input)	--
-- /da function brings up the UI options.	--
----------------------------------------------
	echo("MySlashProcessorFunc")
	InterfaceOptionsFrame_OpenToCategory(coreOpts)
end

--[[
	Real bank containers = -1, 5,6,7,8,9,10,11

	numTabs = GetNumGuildBankTabs()

	
]]
--		/run _GBS:SpoofBank()
--[[
	My test frame to test BANKFRAME_OPENED being fired.]]
	
--[===[@debug@
local testFrame = CreateFrame("Frame")
testFrame:SetScript("OnEvent", function(frame, event, ...)
--~ 	if event == "BANKFRAME_OPENED" then
--~ 		local container = 5
--~ 		Debug("testFrame", event, "slots:"..tostring( _G.GetContainerNumSlots(container) ) )
--~ 		
--~ 		
--~ 		local slots = _G.GetContainerNumSlots(container)
--~ 		local itemLink, count
--~ 		for slot = 1, slots, 1 do
--~ 			itemLink = _G.GetContainerItemLink(container, slot)
--~ 			if itemLink then
--~ 			   _,count = _G.GetContainerItemInfo(container, slot)
--~ 				Debug("testFrame", slot, itemLink, count)
--~ 			end
--~ 		end
--~ 	else
		Debug("testFrame", event, ...)
--~ 	end
end)
--~ testFrame:RegisterEvent("BANKFRAME_OPENED")
--~ testFrame:RegisterEvent("ITEM_LOCK_CHANGED")

--~ testFrame:RegisterEvent("PLAYERBANKSLOTS_CHANGED")
--~ testFrame:RegisterEvent("BAG_UPDATE")
--@end-debug@]===]



function core:ItemLinkToId(itemLink) -- Itemlink, returns the itemID
	local _, _, itemString = itemLink:find("^|c%x+|H(.+)|h%[.+%]")
	if itemString then
		local _, itemId = strsplit(":", itemString)
		return tonumber(itemId)
	end
	return nil
end

local lockEventTimer = false


function core:SetArgs(...)
-- Set global arg1-arg11 values for addons that use them. =/
	for i=1, 11 do 
		_G["arg"..i] = select(i, ...)
	end
end

--- Events
function core:SpoofBank()--		/run _GBS:SpoofBank()
	spoofing = true
	for i,frame in ipairs({GetFramesRegisteredForEvent("BANKFRAME_OPENED")}) do
--~ 		frame:GetScript("OnEvent")
		if frame ~= _G.BankFrame then
			core:SetArgs()
			frame:GetScript("OnEvent")(frame, "BANKFRAME_OPENED")
		end
	end
end

local timerC = 0
local timerS = 0
function core:StartLockEventTimer(container, slot)
	timerC = container
	timerS = slot
	
	self:CancelTimer(lockEventTimer, true)
	lockEventTimer = self:ScheduleTimer("FireLockEventTmr", unmovedLockDelay, "ITEM_LOCK_CHANGED")
end




local ourSpoofingButton = false
local guildSlots = {}
function core:SaveGuildSlots()
	guildSlots = {}--reset
	
	local texture, count, locked, link
	for t=1, GetNumGuildBankTabs() do 
		guildSlots[t] = guildSlots[t] or {}
		for s=1, slotsPerBankTab do 
			link = _G.GetGuildBankItemLink(t, s)
--~ 			if link then
				texture, count, locked =  _G.GetGuildBankItemInfo(t, s)
				guildSlots[t][s] = {
					locked = locked,
					count = count, 
					link = link,
				}
--~ 			else
--~ 				guildSlots[t][s] = false
--~ 			end
		end
	end
	
end

--[[
	Events
		ITEM_LOCK_CHANGED
			Fires whena item is picked up with the slot.
			Fires when a item is dropped with the new location's slot.
			
		GUILDBANK_ITEM_LOCK_CHANGED
			Fires when a item is picked up, we're able to get the location.
			Fires when a item is dropped but can't find the item.
			
			

]]


function core:GUILDBANKFRAME_OPENED(event, ...)
--~ 	Debug(event, "/run _GBS:SpoofBank()")
	if not ourSpoofingButton then
		local f = CreateFrame("Button", nil, _G.GuildBankFrame, "UIPanelButtonTemplate")
		f:SetWidth(60)
		f:SetHeight(21)
		
		f:SetText("Spoof")
		
		f:SetPoint("BOTTOMLEFT", _G.GuildBankFrame, "BOTTOMLEFT", 25, 37)
		
		f:SetScript("OnClick", function(frame, ...)
--~ 			print("Hi!")
			Debug("Spoof Click","Telling other frames your bank has been opened.")
			
			self:CancelTimer(lockEventTimer, true)
			self:SpoofBank()
		end)
		
		ourSpoofingButton = f 
	end
	self:SaveGuildSlots()
--~ 	
end
function core:GUILDBANKFRAME_CLOSED(event, ...)
--~ 	Debug(event, ...)
	
	for i,frame in ipairs({GetFramesRegisteredForEvent("BANKFRAME_CLOSED")}) do
		core:SetArgs()
		frame:GetScript("OnEvent")(frame, "BANKFRAME_CLOSED")
	end
	spoofing = false
	self:CancelTimer(lockEventTimer, true)
end

--[[]]


--[[function core:PLAYERBANKBAGSLOTS_CHANGED(event, ...)
Fires when the bag slots in the bank move. No need to watch this.
--~ 	Debug(event, ...)
end]]

function core:BAG_UPDATE(event, ...)
--[[
	Fires when items in bank bags move.
	Only the container is sent, no slot.
]]
--~ 	local container, slot = ...
--~ 	Debug(event, container, slot)
end

function core:PLAYERBANKSLOTS_CHANGED(event, ...)
--[[ Only fires for main bank pack. not for bags in bank.
		Picking up item
			Nothing happens.
		Droping item in same location.
			Nothing happens.
		Dropping item in new location
			Fires twice for both slots that have changed. 
]]


--~ 	Debug(event, ...)
	
end


function core:GUILDBANKBAGSLOTS_CHANGED(event, ...)
--~ 	Debug(event, ...)
	
	if spoofing == true then
		local texture, count, locked, link
		for t=1, GetNumGuildBankTabs() do 
		
			for s=1, slotsPerBankTab do 
				texture, count, locked =  _G.GetGuildBankItemInfo(t, s)
				link = _G.GetGuildBankItemLink(t, s)
				
				if guildSlots[t][s] then --link and
					if guildSlots[t][s].link ~= link 
					or guildSlots[t][s].locked ~= locked
					or guildSlots[t][s].count ~= count then
	--~ 					Debug(event, link, t, s, "container:"..self:tabToContainer(t))
						if link then
							-- Only fire if there's a item in the slot. The timer will fire the event soon anyway.
							self:CancelTimer(lockEventTimer, true)
							self:FireLockEvent("ITEM_LOCK_CHANGED", self:tabToContainer(t), s, "GUILDBANKBAGSLOTS_CHANGED")
						end
						
						if t == 1 then
							--Slot only
							self:FireLockEvent("PLAYERBANKSLOTS_CHANGED", s)
						else
							--Container only.
							self:FireLockEvent("BAG_UPDATE", self:tabToContainer(t))
						end
						
						
					end
					
					
				end
	
				
			end
		end
	
		self:SaveGuildSlots()
	end
	
	--[[
	if spoofing == true then
		for i,frame in ipairs({GetFramesRegisteredForEvent("PLAYERBANKSLOTS_CHANGED")}) do
			frame:GetScript("OnEvent")(frame, "PLAYERBANKSLOTS_CHANGED")
		end
	end]]
end


function core:GUILDBANK_ITEM_LOCK_CHANGED(event, ...)
--~ 	Debug(event, ...)
	
	if spoofing == true then
		local texture, count, locked, link
		for t=1, GetNumGuildBankTabs() do 
		
			for s=1, slotsPerBankTab do 
				texture, count, locked =  _G.GetGuildBankItemInfo(t, s)
				link = _G.GetGuildBankItemLink(t, s)
	
				-- We only want to fire when a item has been picked up. 
				-- GUILDBANKBAGSLOTS_CHANGED will fire when a item is dropped/moved.
				
				if guildSlots[t][s].locked ~= locked then
					
					if locked then
	--~ 					Debug(event, link, "old:"..tostring(guildSlots[t][s].locked), "new:"..tostring(locked))
						
						self:FireLockEvent("ITEM_LOCK_CHANGED", self:tabToContainer(t), s, "GUILDBANK_ITEM_LOCK_CHANGED")
						
					else
	--~ 					Debug(event, link, "old:"..tostring(guildSlots[t][s].locked), "new:"..tostring(locked), "Fire timer event?")
						core:StartLockEventTimer(self:tabToContainer(t), s)
						
						
					end
					
				end
			
			end
		end
		
		self:SaveGuildSlots()
	end
end


function core:ITEM_LOCK_CHANGED(event, ...)
--~ 	Debug(event, ...)

--~ 	if spoofing == true then
--~ 		self:StartLockEventTimer()
--~ 	end
end


function core:FireLockEventTmr()
--~ 	Debug("FireLockEventTmr", timerC, timerS)
	self:FireLockEvent("ITEM_LOCK_CHANGED", timerC, timerS, "FireLockEventTmr")
end

----------------------------------------------------------------------------------

--[[
	Reagent Restocker uses ITEM_LOCK_CHANGED but if I fire it on GUILDBANK_ITEM_LOCK_CHANGED sometimes items are still locked causing RR to stop in it's tracks.
	So, I'm using a timer. =/
]]
function core:FireLockEvent(event, tab, slot, from)
--~ 	Debug("FireLockEvent", event)
	core:UnregisterEvent(event)
	for i,frame in ipairs({GetFramesRegisteredForEvent(event)}) do
		--No sense firing our own frame or blizzard UI frames. =/
		if	frame ~= core and
			frame ~= _G.BankFrame and
			frame ~= _G.ContainerFrame and
			frame ~= _G.PaperDollItemSlotButton then
			
			self:SetArgs(tab, slot)
			frame:GetScript("OnEvent")(frame, event, tab, slot, from) --self:tabToContainer(tab)
		end
	end
	core:RegisterEvent(event)
end

function core:containerToTab(container)
	if container == BANK_CONTAINER then
		return 1
	end
	if container >= 5 then
		return container - 3
	end
	return nil
end

function core:tabToContainer(tab)
	if tab == 1 then
		return BANK_CONTAINER
	end
	return tab + 3
end

--Hooks
--		/run _GBS:SpoofBank()
function core:GetContainerNumSlots(...)
	local container = ...
	if spoofing == true then
		if _G.GuildBankFrame and _G.GuildBankFrame:IsShown() and container >= 5 then
			local tab = self:containerToTab(container)
	--~ 		Debug("GetContainerNumSlots","Guild bank is open", container, tab, "tabs:"..tostring(GetNumGuildBankTabs()))
			if GetNumGuildBankTabs() >= tab then
				return slotsPerBankTab
			end
		end
	end
	return self.hooks.GetContainerNumSlots(...)
end

function core:GetContainerItemLink(...)
	local container, slot = ...

	if spoofing == true then
		if _G.GuildBankFrame and _G.GuildBankFrame:IsShown() and container >= 5 then
			local tab = self:containerToTab(container)
	--~ 		if _G.GetGuildBankItemLink(tab, slot) then
	--~ 			Debug("GetContainerItemLink", container, slot,"tab:"..tostring(tab), _G.GetGuildBankItemLink(tab, slot))
	--~ 		end
			return _G.GetGuildBankItemLink(tab, slot)
		end
	end
	return self.hooks.GetContainerItemLink(...)
end

--~ texture, count, locked, quality,  readable, lootable, link =  GetContainerItemInfo(container, slot)
--~ texture, count, locked =  GetGuildBankItemInfo(tab, slot)
local lastLink = false

function core:GetContainerItemInfo(...)
	local container, slot = ...
	
	if spoofing == true then
		if _G.GuildBankFrame and _G.GuildBankFrame:IsShown() and container >= 5 then
			local tab = self:containerToTab(container)
			
			
			local itemLink = _G.GetGuildBankItemLink(tab, slot)
	--~ 		if itemLink ~= lastLink then
	--~ 			lastLink = itemLink
	--~ 			Debug("GetContainerItemInfo", ..., itemLink)
	--~ 		end
			Debug("GetContainerItemInfo", container, slot, "tab:"..tostring(tab), itemLink)
	
			if itemLink then
				local texture, count, locked =  _G.GetGuildBankItemInfo(tab, slot)
				local _, _, quality = GetItemInfo(itemLink)
				local readable = nil
				local lootable = nil
				return texture, count, locked, quality, readable, lootable, itemLink
			end
			return nil
		end
	end
	
	return self.hooks.GetContainerItemInfo(...)
end


function core:SplitContainerItem(...)
	local container, slot, amount = ...
	
	if spoofing == true then
		if _G.GuildBankFrame and _G.GuildBankFrame:IsShown() and container >= 5 then
			local tab = self:containerToTab(container)
	--~ 		Debug("SplitContainerItem", "SplitGuildBankItem", tab, slot, amount)
			return _G.SplitGuildBankItem(tab, slot, amount)
		end
	end
	
	return self.hooks.SplitContainerItem(...)
end

function core:PickupContainerItem(...)
	local container, slot = ...

	if spoofing == true then
		if _G.GuildBankFrame and _G.GuildBankFrame:IsShown() and container >= 5 then
			local tab = self:containerToTab(container)
	--~ 		Debug("PickupContainerItem", "PickupGuildBankItem", tab, slot)
			return _G.PickupGuildBankItem(tab, slot)
		end
	end
	
--~ 	Debug("PickupContainerItem", container, slot)
	return self.hooks.PickupContainerItem(...)
end

function core:PutItemInBag(...)
	local container = ...
	
--~ 	if spoofing == true then
--~ 		if _G.GuildBankFrame and _G.GuildBankFrame:IsShown() and container >= 5 then
--~ 			local tab = self:containerToTab(container)
--~ 		end
--~ 	end
	
	return self.hooks.PutItemInBag(...)
end

function core:GetItemCount(...)
	local itemId, includeBank, includeCharges = ...
	
	--[[
	if includeBank ~= false then
		--Since incoming itemID can be a name/itemLink, I wanna make sure I'm getting a link.
		local _, reqItemLink =  GetItemInfo(itemId)
		if reqItemLink then
			itemId = self:ItemLinkToId(reqItemLink)
	
			if _G.GuildBankFrame and _G.GuildBankFrame:IsShown() then
				local totalCount = self.hooks.GetItemCount(itemId, false, true) or 0
				local itemLink, itemID, count
				for t=1, GetNumGuildBankTabs() do 
				
					for s=1, slotsPerBankTab do 
						itemLink = _G.GetGuildBankItemLink(t, s)
						if itemLink then
							itemID = self:ItemLinkToId(itemLink)
							
--~ 							Debug("GetItemCount B", t, s, itemLink, itemID, itemId)
							
							if itemID == itemId then
								
								_, count =  _G.GetGuildBankItemInfo(t, s)
								totalCount = totalCount + count
								
								Debug("GetItemCount C", t, s, itemLink, count, totalCount)
							end
							
							
						end
					end
					

				end
				Debug("GetItemCount D", reqItemLink, totalCount)
				return totalCount
			end
		end
	end]]

	return self.hooks.GetItemCount(...)
end

--[[function core:UseContainerItem(...)
	local container, slot = ...
end]]


function core:GetNumBankSlots(...)
	--[[
		Number of bag slots purchaced. 
	]]
	
	if spoofing == true then
		if _G.GuildBankFrame and _G.GuildBankFrame:IsShown() then
			return GetNumGuildBankTabs(), 1
		end
	end
	
	return self.hooks.GetNumBankSlots(...)
end



function core:GetInventoryItemLink(...)
--~ 	Debug("GetInventoryItemLink", self.hooks.GetInventoryItemLink(...))
	
	if spoofing == true then
		local unitID, slot = ...
		if UnitIsUnit(unitID, "player") then
			if slot >= 68 and slot <= 74 then
				local name, link = GetItemInfo(fakeBagID) --Portable Hole
				return link
			end
		end
	end
	
	return self.hooks.GetInventoryItemLink(...)
end

function core:GetInventoryItemTexture(...)
--~ 	Debug("GetInventoryItemTexture", self.hooks.GetInventoryItemTexture(...))

	if spoofing == true then
		local unitID, slot = ...
		if UnitIsUnit(unitID, "player") then
			if slot >= 68 and slot <= 74 then
				local _, _, _, _,  _, _, _, _,  _, texture =  GetItemInfo(fakeBagID)
				return texture
			end
		end
	end
	return self.hooks.GetInventoryItemTexture(...)
end
